
################################################################

# Description of a SOAP type 
# Later in the hierarcahy, we include a function that can convert to and from a SOAP representation and R.

setClass("GenericSchemaType", representation(name="character",
                                    ns="character",
                                    nsuri="character",
                                    documentation = "character"))


setClass("SOAPType", representation(count = "numeric", # use numeric rather than integer because of Inf for unbounded1.
                                    abstract = "logical"
                                    ), contains = "GenericSchemaType")


# For representing arbitrary content, e.g. <complexType/>
setClass("AnySOAPType", contains = "SOAPType", prototype = list(name = "AnySOAPType"))

setClass("SOAPVoidType", representation("SOAPType"),
            prototype = list(name = "null", ns = "xsi"))
setClass("SOAPAnyType", representation("SOAPType"))

setClass("SOAPTypeReference", representation("SOAPType"))

setClass("TerminalSOAPType", representation("VIRTUAL"))

setClass("BasicSOAPType", representation("SOAPType",
                                         fromConverter = "function",  # From SOAP to R
                                         toConverter = "function"))   # to SOAP from R

setClassUnion("SOAPTypeOrList", c("SOAPType", "list", "NULL"))

setClass("SOAPComplexType",
          representation(attributes = "list", content = "SOAPTypeOrList", xmlAttrs = "character"),
          contains = "BasicSOAPType")
            


#XXX Is this a setClassUnion ?
setClass("PrimitiveSOAPType", contains = c("BasicSOAPType", "TerminalSOAPType"))


setClass("SOAPDate", contains= "PrimitiveSOAPType")
setClass("SOAPDateTime", contains= "PrimitiveSOAPType")
setClass("SOAPTime", contains= "PrimitiveSOAPType")

setClass("gYear", contains= "integer") # "PrimitiveSOAPType")


setClass("SimpleSequenceType", representation("BasicSOAPType", elementType = "character", elType = "SOAPType"))

# White space spearated values
setClass("ListType", contains = "SimpleSequenceType")
setClass("RestrictedListType", representation(elements = "character"),  # is elements somehow already in elType and elementType
                               contains = "ListType")

                                              
setClass("ArrayType", representation("SimpleSequenceType", dims = "integer"))



# A definition for a complex data type defined in a WSDL file.
setClass("CompositeTypeDefinition", representation("BasicSOAPType",  slotTypes = "list"), contains = "TerminalSOAPType")
setClass("ClassDefinition",  contains = "CompositeTypeDefinition")
setClass("UnionDefinition", contains = "CompositeTypeDefinition")
setClass("ArrayClassDefinition",  contains = "CompositeTypeDefinition")

#setClass("EnumValuesDef",  contains = "character")
setClass("EnumValuesDef",  representation(values = "character"), contains = "SOAPType")

setClass("RestrictedSetInteger", representation(values = "integer"), contains = "BasicSOAPType")

setClass("ExtendedClassDefinition", representation(base = "character"), contains = "ClassDefinition")

# Was SOAPType and not BasicSOAPType. But needed the fromConverter. See 
#   http://www.weather.gov/forecasts/xml/DWMLgen/wsdl/ndfdXML.wsdl
setClass("RestrictedStringType", contains = "BasicSOAPType")
setClass("RestrictedStringDefinition", representation(values = "character"), contains = "RestrictedStringType")
setClass("RestrictedStringPatternDefinition", representation(pattern = "character"), contains = "RestrictedStringType")

#
# When we generate an S4 class to represent a SOAP compound data type,
# we have it extend this degenerate class. This allows us to
# provide methods that dispatch on that class such as converting from
# S to SOAP values and vice-verse.
setClass("VirtualXMLSchemaClass")




#XX need validity methods to ensure it is a proper restriction of a character.
setClass("token", contains = "character")
#XXX we won't leave this as a character
setClass("duration", contains = "character")

setClass("URIString", contains = "URI")

setOldClass("WSDLTypeDescription")

# A description of a SOAP method in a server.
# See http://www-128.ibm.com/developerworks/webservices/library/ws-whichwsdl/ for
# information about bindingStyle and use fields.





